__all__ = ["asarray", "asset", "astree"]

# standard library
from collections.abc import Hashable
from typing import Any

# dependencies
import pandas as pd
import xarray as xr
from .spec import parse


def asarray(obj: Any, /) -> xr.DataArray:
    """Convert an xarray specification to an xarray DataArray."""
    specs = parse(obj)
    da = xr.DataArray(
        data=next(iter(to_vars(specs).values())),
        coords=to_coords(specs),
    )
    da.attrs.update(to_attrs(specs))
    return da


def asset(obj: Any, /) -> xr.Dataset:
    """Convert an xarray specification to an xarray Dataset."""
    specs = parse(obj)
    ds = xr.Dataset(
        data_vars=to_vars(specs),
        coords=to_coords(specs),
    )
    ds.attrs.update(to_attrs(specs))
    return ds


def astree(obj: Any, /) -> xr.DataTree:
    """Convert an xarray specification to an xarray DataTree."""
    specs = parse(obj)
    nodes: dict[str, xr.Dataset] = {}

    for node, subspecs in specs.groupby("xarray_node"):
        ds = xr.Dataset(
            data_vars=to_vars(subspecs),
            coords=to_coords(subspecs),
        )
        ds.attrs.update(to_attrs(subspecs))
        nodes[node] = ds  # type: ignore

    return xr.DataTree.from_dict(nodes)  # type: ignore


def to_attrs(specs: pd.DataFrame, /) -> dict[Any, Any]:
    """Convert a specification DataFrame to xarray attributes."""
    attrs: dict[Any, Any] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_use == "attr":
            attrs[spec.xarray_name] = spec.data
        elif spec.xarray_use == "attrs":
            attrs.update(spec.data)

    return attrs


def to_coords(specs: pd.DataFrame, /) -> dict[Hashable, xr.DataArray]:
    """Convert a specification DataFrame to xarray coordinates."""
    coords: dict[Hashable, xr.DataArray] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_use == "coord":
            coords[spec.xarray_name] = xr.DataArray(
                data=spec.data,
                dims=spec.xarray_dims,
                name=spec.xarray_name,
                attrs=spec.xarray_attrs,
            ).astype(  # type: ignore
                spec.xarray_dtype,
                copy=False,
            )
        elif spec.xarray_use == "coords":
            for name, data in spec.data.items():
                coords[name] = xr.DataArray(
                    data=data,
                    dims=spec.xarray_dims,
                    name=name,
                    attrs=spec.xarray_attrs,
                ).astype(  # type: ignore
                    spec.xarray_dtype,
                    copy=False,
                )

    return coords


def to_vars(specs: pd.DataFrame, /) -> dict[Hashable, xr.DataArray]:
    """Convert a specification DataFrame to xarray data variables."""
    vars: dict[Hashable, xr.DataArray] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_use == "data":
            vars[spec.xarray_name] = xr.DataArray(
                data=spec.data,
                dims=spec.xarray_dims,
                name=spec.xarray_name,
                attrs=spec.xarray_attrs,
            ).astype(  # type: ignore
                spec.xarray_dtype,
                copy=False,
            )
        elif spec.xarray_use == "vars":
            for name, data in spec.data.items():
                vars[name] = xr.DataArray(
                    data=data,
                    dims=spec.xarray_dims,
                    name=name,
                    attrs=spec.xarray_attrs,
                ).astype(  # type: ignore
                    spec.xarray_dtype,
                    copy=False,
                )

    return vars



# dependencies
import xarrayspecs


def test_namespace() -> None:
    # submodules
    xarrayspecs.api
    xarrayspecs.spec
    # aliases
    xarrayspecs.Dims
    xarrayspecs.Dtype
    xarrayspecs.asarray
    xarrayspecs.asset
    xarrayspecs.attrs
    xarrayspecs.dims
    xarrayspecs.dtype
    xarrayspecs.name
    xarrayspecs.node
    xarrayspecs.parse
    xarrayspecs.use



__all__ = ["Dims", "Dtype", "attrs", "dims", "dtype", "name", "node", "parse", "use"]

# standard library
from collections.abc import Hashable, Iterable
from typing import Annotated, Any, Literal, TypeVar

# dependencies
import pandas as pd
from typespecs import ITSELF, Spec, SpecFrame, from_annotated
from typing_extensions import get_args, get_origin

# type hints
T = TypeVar("T")


Dims = Annotated[T, Spec(xarray_dims=ITSELF)]
"""Type hint for xarray dimensions."""


Dtype = Annotated[T, Spec(xarray_dtype=ITSELF)]
"""Type hint for xarray data type."""


Use = Literal["attr", "attrs", "coord", "coords", "data", "vars"]
"""Type hint for xarray use."""


def attrs(attrs: dict[Any, Any] | None, /) -> Spec:
    """Returns a type specification for xarray attributes."""
    return Spec(xarray_attrs=attrs)


def dims(dims: Iterable[str] | str | None, /) -> Spec:
    """Returns a type specification for xarray dimensions."""
    return Spec(xarray_dims=dims)


def dtype(dtype: Any | None, /) -> Spec:
    """Returns a type specification for xarray data type."""
    return Spec(xarray_dtype=dtype)


def name(name: Hashable | None, /) -> Spec:
    """Returns a type specification for xarray name."""
    return Spec(xarray_name=name)


def node(node: str | None, /) -> Spec:
    """Returns a type specification for xarray node."""
    return Spec(xarray_node=node)


def use(use: Use, /) -> Spec:
    """Returns a type specification for xarray use."""
    return Spec(xarray_use=use)


def parse(obj: Any, /) -> SpecFrame:
    """Returns a specification DataFrame for xarray."""
    specs = from_annotated(
        obj,
        default={
            "xarray_attrs": None,
            "xarray_dims": None,
            "xarray_dtype": None,
            "xarray_name": None,
            "xarray_node": "/",
            "xarray_use": pd.NA,
        },
    )

    index = specs.index.to_series()
    specs["xarray_dims"] = specs["xarray_dims"].apply(to_dims)
    specs["xarray_name"] = specs["xarray_name"].fillna(index)
    return specs


def to_dims(obj: Any, /) -> tuple[str, ...] | None:
    """Convert a dimensions-like object to dimensions."""
    if obj is None:
        return None

    try:
        return (to_str(obj),)
    except TypeError:
        return tuple(map(to_str, to_tuple(obj)))


def to_str(obj: Any, /) -> str:
    """Convert a string-like object to a string."""
    if get_origin(obj) is Literal:
        return str(get_args(obj)[0])

    if isinstance(obj, str):
        return str(obj)

    raise TypeError(f"Cannot convert {obj!r} to string.")


def to_tuple(obj: Any, /) -> tuple[Any, ...]:
    """Convert a tuple-like object to a tuple."""
    if get_origin(obj) is tuple:
        return tuple(get_args(obj))

    if isinstance(obj, Iterable):
        return tuple(obj)  # type: ignore

    raise TypeError(f"Cannot convert {obj!r} to tuple.")



__all__ = [
    # submodules
    "api",
    "spec",
    # aliases
    "Dims",
    "Dtype",
    "asarray",
    "asset",
    "astree",
    "attrs",
    "dims",
    "dtype",
    "name",
    "node",
    "parse",
    "use",
]
__version__ = "0.1.0"


# dependencies
from . import api, spec
from .api import *
from .spec import *
